from __future__ import absolute_import, unicode_literals

import logging
import os
import platform
import ssl
from xml.etree import ElementTree

import requests
from six.moves.urllib.parse import urljoin

from altapay import (__api_base_url__, __github_url__, __version__, exceptions,
                     utils)
from altapay.resource import Resource

logger = logging.getLogger(__name__)


class API(object):
    details = 'requests {requests}; python {python}; {ssl}'.format(
        requests=requests.__version__, python=platform.python_version(),
        ssl=ssl.OPENSSL_VERSION)
    user_agent = 'coolshop-com/AltaPay/{version} (+{url}; {details})'.format(
        version=__version__, url=__github_url__, details=details)
    _is_authenticated = False

    def __init__(self, **kwargs):
        """
        Instantiate an AltaPay API object.
        """
        auto_login = kwargs.get('auto_login', True)

        self.mode = kwargs.get('mode', 'test')
        self.url = kwargs.get('url', self._default_url)

        # If production mode, shop_name must be passed through kwargs, as it
        # is neede for the URL
        self.shop_name = kwargs.get('shop_name', '')
        if self.mode == 'production' and not self.shop_name:
            raise exceptions.APIError(
                'No shop_name was provided, even though production mode was '
                'specified. When running in production, a shop_name is ',
                'needed, as this is used to produce the AltaPay URL.')
        self.url = self.url.format(shop_name=self.shop_name)

        # If account information is passed through the kwargs, use these
        # If not, attempt to read from the environment
        self.account = kwargs.get('account', '')
        self.password = kwargs.get('password', '')

        if not self.account or not self.password:
            self.account = os.environ.get('ALTAPAY_ACCOUNT_NAME', '')
            self.password = os.environ.get('ALTAPAY_ACCOUNT_PASSWORD', '')

        if not self.url:
            raise exceptions.APIError(
                'No API URL was provided, and the selected mode could not be '
                'mapped to a default URL: ' + self.mode)

        if auto_login:
            self.login()

    @property
    def _default_url(self):
        return __api_base_url__.get(self.mode, '')

    @property
    def _auth(self):
        return (self.account, self.password)

    def login(self):
        """
        Validates the account name and password against the AltaPay service.
        This method should always be called before attempting any other calls,
        and is automatically called once the :py:class:`altapay.api.API`
        object is instantiated, unless explictly disabled.

        Raises:
            UnauthorizedAccessError: if the supplied credentials are not valid.
        """
        if self._is_authenticated:
            return

        if Resource.create_from_response(self.get('API/login')).success:
            self._is_authenticated = True
            return

        # Nothing at this point is intentional; if the login was not
        # successful, the HTTP response code will be 401 and result in an
        # UnauthorizedError from _response()

    def test_connection(self):
        """
        Tests the connection to the AltaPay service. This operation does
        not require valid API credentials, and as such can only be used to
        assert if AltaPay is responding.

        :rtype: :samp:`True` if a valid response is returned, otherwise
            :samp:`False`.
        """
        return Resource.create_from_response(self.get('API/testConnection'))

    def _headers(self):
        return {
            'User-Agent': self.user_agent
        }

    def _request(self, url, method, params={}, headers={}):
        logger.debug('Mode: ' + self.mode)
        logger.debug('URL: ' + url)
        logger.debug('Method: ' + method)
        logger.debug('Params: ' + str(params))
        logger.debug('Headers: ' + str(headers))
        logger.debug('Is authenticated: ' + str(self._is_authenticated))

        response = requests.request(
            method, url, params=params, headers=headers, auth=self._auth)

        return self._response(response, response.content.decode('utf-8'))

    def _response(self, response, content):
        status = response.status_code

        logger.debug('Status: ' + str(status))
        logger.debug('Content: ' + content)

        if status in (200, 201):
            return utils.etree_to_dict(
                ElementTree.XML(content.encode('utf-8')))
        elif status == 401:
            raise exceptions.UnauthorizedAccessError(
                'Credentials could not be validated against the AltaPay '
                'service.')
        elif 500 <= status <= 599:
            raise exceptions.ServerError(
                'AltaPay service server error. Response code was: {}'
                .format(status))

        raise exceptions.ResponseStatusError(
            'Response code not allowed: {status}'.format(status=status))

    def get(self, resource, parameters={}, headers={}):
        """
        Perform a GET HTTP request on a resource.

        :arg resource: the resource to GET
        :arg parameters: a dictionary of GET parameters for the resource
        :arg headers: optional headers. If specified, these will override the
            default headers.

        :returns:
            A response from the AltaPay service as a :samp:`dict`.

        :raises:
            :UnauthorizedAccessError: If the supplied credentials are not
                valid.

            :ResponseStatusError: If the response code from AltaPay is not a
                subset of the allowed response codes.

        """
        return self._request(
            urljoin(self.url, resource), 'GET',
            params=utils.http_build_query(parameters),
            headers=headers or self._headers())



from __future__ import absolute_import, unicode_literals

from altapay.resource import Resource


class Payment(Resource):
    def create(self, terminal, shop_orderid, amount, currency, **kwargs):
        """
        Create a payment request.

        :arg terminal: name of the targeted AltaPay terminal
        :arg shop_orderid: your order ID to be attached to the payment resource
        :arg amount: order amount in floating point
        :arg currency: currency for the payment resource
        :arg \*\*kwargs: used for remaining, optional, payment request
            parameters, see the AltaPay documentation for a full list.
            Note that you will need to use lists and dictionaries to map the
            URL structures from the AltaPay documentation into these kwargs.

        :rtype: :samp:`True` if a payment was created, otherwise :samp:`False`.
        """
        parameters = {
            'terminal': terminal,
            'shop_orderid': shop_orderid,
            'amount': amount,
            'currency': currency,
        }

        parameters.update(kwargs)

        response = self.api.get(
            'API/createPaymentRequest', parameters=parameters)
        self.merge_response(response)
        return self.success



from __future__ import absolute_import, unicode_literals

from xml.etree import ElementTree

import altapay.transaction
from altapay import utils
from altapay.resource import Resource


class Callback(Resource):
    def transactions(self, **kwargs):
        """
        List all of the transactions returned by the callback.

        :param auth_type: the authentication type you wish to filter.
            Defaults to empty string, which means no filter will be made.

        :rtype: List of :py:class:`altapay.Transaction` objects.
        """
        def _check_value(element, **kwargs):
            for key, value in element.items():
                check_key = utils.to_pythonic_name(key)
                check_value = kwargs.pop(check_key, '')
                if check_value and element[key] != check_value:
                    return False
            return True

        data = self.__data__['transactions']['transaction']

        if not isinstance(data, list):
            data = [data]

        transaction_set = data

        transaction_set = [
            transaction for transaction in data
            if _check_value(transaction, **kwargs)
        ]

        return [
            altapay.transaction.Transaction(
                self.version, self.__header__, transaction)
            for transaction in transaction_set]

    @classmethod
    def create_invoice_reservation(cls, terminal, shop_orderid, amount,
                                   currency, api, **kwargs):
        """
        Create a new invoice without first creating a payment.

        :rtype: :py:class:`altapay.Transaction`
        """
        parameters = {
            'terminal': terminal,
            'shop_orderid': shop_orderid,
            'amount': amount,
            'currency': currency
        }

        parameters.update(kwargs)

        response = api.get(
            'API/createInvoiceReservation', parameters=parameters
        )['APIResponse']

        return cls(
            response['@version'], response['Header'], response['Body'])

    @classmethod
    def from_xml_callback(cls, callback):
        """
        Instantiate a :py:class:`altapay.Callback` object from an XML response.

        :rtype: :py:class:`altapay.Callback` instance.
        """
        if isinstance(callback, str):
            callback = ElementTree.XML(callback.encode('utf-8'))
            response = utils.etree_to_dict(callback)['APIResponse']
        elif isinstance(callback, ElementTree.Element):
            response = utils.etree_to_dict(callback)['APIResponse']
        else:
            response = callback

        return cls(
            response['@version'], response['Header'],
            response['Body'])



from __future__ import absolute_import, unicode_literals

import altapay.callback
from altapay import exceptions
from altapay.resource import Resource


class Transaction(Resource):
    @classmethod
    def find(cls, transaction_id, api):
        """
        Find exactly one transaction by a transaction ID.

        :param transaction_id: ID of the transaction in AltaPay
        :param api: An API object which will be used for AltaPay communication.

        :rtype: :py:class:`altapay.Transaction`
        """
        response = api.get(
            'API/payments', parameters={'transaction_id': transaction_id}
        )['APIResponse']

        try:
            transaction = response['Body']['Transactions']['Transaction']
        except KeyError:
            raise exceptions.ResourceNotFoundError(
                'No Transaction found matching transaction ID: {}'.format(
                    transaction_id))

        if isinstance(transaction, list):
            raise exceptions.MultipleResourcesError(
                'More than one Payment was found. Total found is: {}'.format(
                    len(response['Body']['Transactions'])))

        return cls(
            response['@version'], response['Header'], transaction, api=api)

    def capture(self, **kwargs):
        """
        Capture a reservation on a transaction.

        :param \*\*kwargs: used for optional capture parameters, see the
            AltaPay documentation for a full list.
            Note that you will need to use lists and dictionaries to map the
            URL structures from the AltaPay documentation into these kwargs.
        """
        parameters = {
            'transaction_id': self.transaction_id
        }

        parameters.update(kwargs)

        response = self.api.get(
            'API/captureReservation', parameters=parameters)['APIResponse']

        return Transaction(
            response['@version'], response['Header'],
            response['Body']['Transactions']['Transaction'], api=self.api)

    def charge_subscription(self, **kwargs):
        """
        This will charge a subscription using a capture. Can be called many
        times on a subscription.

        If amount is not sent as an optinal parameter, the amount specified in
        the original setup of the subscription will be used.

        :param \*\*kwargs: used for optional charge subscription parameters,
            see the AltaPay documentation for a full list.
            Note that you will need to use lists and dictionaries to map the
            URL structures from the AltaPay documentation into these kwargs.

        :rtype: :py:class:`altapay.Callback` object.
        """
        parameters = {
            'transaction_id': self.transaction_id
        }

        parameters.update(kwargs)

        response = self.api.get(
            'API/chargeSubscription', parameters=parameters)['APIResponse']

        return altapay.callback.Callback.from_xml_callback(response)

    def reserve_subscription_charge(self, **kwargs):
        """
        This will create a reservation on a subscription. Can be called many
        times on a subscription.

        If amount is not sent as an optinal parameter, the amount specified in
        the original setup of the subscription will be used.

        :param \*\*kwargs: used for optional reserve subscription parameters,
            see the AltaPay documentation for a full list.
            Note that you will need to use lists and dictionaries to map the
            URL structures from the AltaPay documentation into these kwargs.

        :rtype: :py:class:`altapay.Callback` object.
        """
        parameters = {
            'transaction_id': self.transaction_id
        }

        parameters.update(kwargs)

        response = self.api.get(
            'API/reserveSubscriptionCharge',
            parameters=parameters)['APIResponse']

        return altapay.callback.Callback.from_xml_callback(response)

    def release(self):
        """
        This will release the reservation on the transaction. This is useful
        if you for whatever reason do not want to capture the payment.

        Refer to the AltaPay documentation for edge cases surround this method.

        :rtype: :py:class:`altapay.Callback` object.
        """
        parameters = {
            'transaction_id': self.transaction_id
        }

        response = self.api.get(
            'API/releaseReservation', parameters=parameters)['APIResponse']

        return altapay.callback.Callback.from_xml_callback(response)



from __future__ import absolute_import, unicode_literals

import re
from collections import OrderedDict, defaultdict
from xml.etree import ElementTree

from six import text_type
from six.moves.urllib.parse import urlencode


def to_pythonic_name(name):
    """
    Create a Pythonic version of a string.

    *Note: This is an internal API and may be changed without notice.*

    :arg name: string to build a Pythonic version of.

    :rtype: string
    """
    s1 = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', name)
    return re.sub('([a-z0-9])([A-Z])', r'\1_\2', s1).lower()


def to_pythonic_dict(dictionary):
    """
    *Note: This is an internal API and may be changed without notice.*
    """
    # TODO: This should probably be cleaned up, and also support dictionaries
    # in lists
    return {
        to_pythonic_name(k): (
            to_pythonic_dict(v) if isinstance(v, dict) else v)
        for k, v in dictionary.items()
    }


def handle_xml_value(value):
    """
    The AltaPay XML does not contain a scheme, and as such, guesswork has to be
    employed in order to produce decent values.

    This function parses values of the decoded XML, and ensures both digits and
    boolean values.

    *Note: This is an internal API and may be changed without notice.*

    :arg value: value to be parsed (can be a complex datatype)

    :rtype: depends on the input argument
    """
    if not isinstance(value, str) and not isinstance(value, text_type):
        return value

    value = text_type(value)

    if value == 'true':
        return True
    elif value == 'false':
        return False
    elif value.isdigit():
        return int(value)

    try:
        return float(value)
    except ValueError:
        pass

    return value


def xml_to_dict(xml):
    """
    :arg xml: XML document in string representation

    :rtype: :samp:`dict`
    """
    return etree_to_dict(ElementTree.XML(xml.encode('utf-8')))


def etree_to_dict(tree):
    """
    *Note: This is an internal API and may be changed without notice.*
    """
    d = {tree.tag: {} if tree.attrib else None}
    children = list(tree)
    if children:
        dd = defaultdict(list)
        for dc in map(etree_to_dict, children):
            for k, v in dc.items():
                dd[k].append(handle_xml_value(v))
        d = {
            tree.tag: {
                k: handle_xml_value(v[0])
                if len(v) == 1 else handle_xml_value(v)
                for k, v in dd.items()
            }
        }
    if tree.attrib:
        d[tree.tag].update(
            ('@' + k, handle_xml_value(v)) for k, v in tree.attrib.items())
    if tree.text:
        text = tree.text.strip()
        if children or tree.attrib:
            if text:
                d[tree.tag]['#text'] = text
        else:
            d[tree.tag] = text
    return d


def http_build_query(payload):
    """
    Build a query string that matches the way PHP does it with
    :samp:`http_build_query`.

    In output, this function loosely matches what PHP does in the function
    :samp:`http_build_query`. It handles complex types of both dict and list.

    If :py:class:`collections.OrderedDict` is used, the order of the keys will
    be preserved in the finalized query string.

    *Note: This is an internal API and may be changed without notice.*

    :arg payload: the payload to convert to a query string. This has to be
        :samp:`dict` compatible, but can hold lists as values in the
        dictionary. Nested dictionaries can be used, and lists can hold
        dictionaries.

    :rtype: :samp:`string` that can be used as a GET parameter for HTTP
        requests
    """
    # Here be dragons
    def unpack_dict(value, key=None):
        if isinstance(value, list):
            for i, elem in enumerate(value):
                yield list(unpack_dict(elem, key='{}[{}]'.format(key, i)))
        elif isinstance(value, dict):
            for k, v in value.items():
                yield list(unpack_dict(v, key='{}[{}]'.format(key, k)))
        else:
            yield key, text_type(value)

    def unpack_list(value, data):
        for elem in value:
            if isinstance(elem, list):
                unpack_list(elem, data)
            else:
                data[elem[0]] = elem[1]

    data = OrderedDict()
    for key, value in payload.items():
        unpack_list(list(unpack_dict(value, key)), data)

    return urlencode(data)



from __future__ import absolute_import, unicode_literals


class AltaPayException(Exception):
    """
    Generic exception class for the AltaPay SDK. All specific exceptions raised
    from the SDK will inherit from this exceptions.
    """


class APIError(AltaPayException):
    """
    Raised if the API object could not be configured with the given parameters.
    """


class UnauthorizedAccessError(AltaPayException):
    """
    Raised on unauthorized errors against the AltaPay service.

    Corresponds to HTTP status code 401.
    """


class ResponseStatusError(AltaPayException):
    """
    The response carried out against the AltaPay service did not respond with
    the expected response status code.
    """


class ServerError(AltaPayException):
    """
    Raised when an error in the 500 range is returned from the AltaPay service.
    """


class MultipleResourcesError(AltaPayException):
    """
    Raised if more than one Resource was found when attempting to look up a
    single resource.
    """


class ResourceNotFoundError(AltaPayException):
    """
    Raised if a resource is not found.
    """



from __future__ import absolute_import, unicode_literals

from altapay import utils


class Resource(object):
    """
    Base class that maps an AltaPay response into a Python like representation.
    """
    def __init__(self, version=None, header=None, body=None, api=None):
        self.__dict__['api'] = api

        super(Resource, self).__setattr__('version', version)
        super(Resource, self).__setattr__('__header__', header or {})
        super(Resource, self).__setattr__('__data__', {})

        self.merge(body or {})

    @classmethod
    def create_from_response(cls, response):
        """
        Instantiate a new :py:class:`altapay.Resource` object from
        a response.

        :arg response: a response of type :samp:`dict`. The response most
            conform to the AltaPay response format, which means it must carry
            four keys called :samp:`@version`, :samp:`APIResponse`,
            :samp:`Header` and :samp:`Body`.

        :rtype: a new version object of type
            :py:class:`altapay.Resource`.
        """
        try:
            api_response = response['APIResponse']
            header = utils.to_pythonic_dict(api_response['Header'])
            body = api_response['Body']
            return cls(
                version=api_response['@version'], header=header, body=body)
        except KeyError:
            raise ValueError('XML document was not in expected format')

    def __str__(self):
        return self.__data__.__str__()

    def __repr__(self):
        return self.__data__.__str__()

    def __getattr__(self, name):
        try:
            return self.__data__[name]
        except KeyError:
            return super(Resource, self).__getattribute__(name)

    def __setattr__(self, name, value):
        self.__data__[name] = value

    def __contains__(self, name):
        return name in self.__data__

    def merge(self, attributes):
        for key, value in attributes.items():
            setattr(
                self, utils.to_pythonic_name(key),
                utils.to_pythonic_dict(value)
                if isinstance(value, dict) else value)

    def merge_response(self, response):
        try:
            api_response = response['APIResponse']
            super(Resource, self).__setattr__(
                'version', api_response['@version'])
            super(Resource, self).__setattr__(
                '__header__', utils.to_pythonic_dict(api_response['Header']))
            self.merge(api_response['Body'] or {})
        except KeyError:
            raise ValueError('XML document was not in the expected format')

    @property
    def error(self):
        error_code = int(self.__header__.get('error_code', 0))
        error_message = self.__header__.get('error_message', '')

        return {
            'code': error_code,
            'message': error_message
        }

    @property
    def success(self):
        return self.error['code'] == 0



import re

from setuptools import setup, find_packages


def get_version():
    with open('altapay/__init__.py') as f:
        return re.search(
            r'^__version__\s*=\s*[\'"]([^\']*)[\'"]', f.read(),
            re.MULTILINE).group(1)


def get_url():
    with open('altapay/__init__.py') as f:
        return re.search(
            r'^__github_url__\s*=\s*[\'"]([^\']*)[\'"]', f.read(),
            re.MULTILINE).group(1)

version = get_version()
github_url = get_url()

with open('README.rst', 'r') as f:
    readme = f.read()

with open('CHANGELOG.rst', 'r') as f:
    changelog = f.read()

requires = [
    'requests',
    'six'
]

setup(
    name='altapay',
    version=version,
    url=github_url,
    license='MIT',
    description='Unofficial Python SDK for AltaPay (formerly Pensio).',
    long_description=readme + '\n\n' + changelog,
    author='Coolshop.com',
    author_email='altapaysdk@coolshop.com',
    packages=find_packages(where='.', exclude=('tests*',)),
    install_requires=requires,
    classifiers=[
        'Development Status :: 1 - Planning',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 3'
    ])



__title__ = 'altapay'
__version__ = '1.0'
__author__ = 'Coolshop.com'
__license__ = 'MIT'
__github_url__ = 'https://github.com/coolshop-com/AltaPay'
__api_base_url__ = {
    'test': 'https://testgateway.altapaysecure.com/merchant/',
    'production': 'https://{shop_name}.altapaysecure.com/merchant/'
}

from altapay.api import API  # NOQA
from altapay.callback import Callback  # NOQA
from altapay.payment import Payment  # NOQA
from altapay.resource import Resource  # NOQA
from altapay.transaction import Transaction  # NOQA
